import logging
import os
import re
import subprocess

import pathlib

import typer

from portal_tool.installer.repo.build_models import ConfigurePreset
from portal_tool.installer.configurators.configurator import (
    Configurator,
    CompilerDetails,
)


class WindowsConfigurator(Configurator):
    def __init__(self, yes: bool):
        logging.info("Running Windows 11 configurator")

    def _try_install_vcpkg_dependencies(self) -> None:
        pass

    def _install_package(self, packages: list[str]) -> None:
        raise NotImplementedError

    def validate_compilers(self) -> list[CompilerDetails]:
        typer.echo("Validating compilers...")

        clang_valid = False
        msvc_valid = False

        found_compilers = []

        try:
            result = subprocess.run(
                ["clang", "--version"], capture_output=True, text=True, timeout=5
            )
            if result.returncode == 0:
                match = re.search(r"clang version (\d+)\.(\d+)", result.stdout)
                installation_path = re.search("InstalledDir: (.*)", result.stdout)
                if match:
                    major = int(match.group(1))
                    if major >= 19:
                        path_info = (
                            f" ({installation_path.group(1)})"
                            if installation_path
                            else ""
                        )
                        typer.echo(f"Clang {major}.{match.group(2)} found{path_info}")
                        clang_valid = True

                        clang_result = subprocess.run(
                            ["where.exe", "clang"],
                            capture_output=True,
                            text=True,
                            timeout=5,
                        )
                        if clang_result.returncode != 0:
                            typer.echo(
                                "clang executable exists in PATH but failed to fetch exe location"
                            )
                            raise FileNotFoundError
                        clang_location = clang_result.stdout.strip()

                        clangpp_result = subprocess.run(
                            ["where.exe", "clang++"],
                            capture_output=True,
                            text=True,
                            timeout=5,
                        )
                        if clangpp_result.returncode != 0:
                            typer.echo(
                                "clang++ executable exists in PATH but failed to fetch exe location"
                            )
                            raise FileNotFoundError
                        clangpp_location = clangpp_result.stdout.strip()

                        found_compilers.append(
                            CompilerDetails(
                                name="clang",
                                c_compiler=clang_location,
                                cpp_compiler=clangpp_location,
                            )
                        )
                    else:
                        typer.echo(
                            f"Clang {major}.{match.group(2)} found, but version 19+ is required"
                        )
        except (subprocess.SubprocessError, FileNotFoundError):
            typer.echo("Clang not found")

        try:
            program_files_x86_path = os.environ.get("ProgramFiles(x86)", "")
            vs_where_path = (
                pathlib.Path(program_files_x86_path)
                / "Microsoft Visual Studio"
                / "Installer"
                / "vswhere.exe"
            )

            result = subprocess.run(
                [vs_where_path], capture_output=True, text=True, timeout=5
            )

            output = result.stdout + result.stderr
            match = re.search(r"installationVersion: (\d+)\.(\d+).(\d+)\.(\d+)", output)
            installation_path = re.search("installationPath: (.*)", output)
            if match:
                major = int(match.group(1))
                minor = int(match.group(2))
                version_str = f"{major}.{minor}"
                if major >= 17:
                    path_info = (
                        f" ({installation_path.group(1)})" if installation_path else ""
                    )
                    typer.echo(f"MSVC {version_str} found{path_info}")
                    msvc_valid = True

                    vs_install_path = pathlib.Path(installation_path.group(1).strip())
                    msvc_tools_dir = vs_install_path / "VC" / "Tools" / "MSVC"
                    if not msvc_tools_dir.exists():
                        typer.echo("MSVC tools directory not found")
                        raise FileNotFoundError
                    toolset_versions = os.listdir(msvc_tools_dir)
                    if not toolset_versions:
                        typer.echo("No MSVC toolset versions found")
                        raise FileNotFoundError
                    toolset_version = sorted(toolset_versions)[-1]
                    cl_path = (
                        msvc_tools_dir
                        / toolset_version
                        / "bin"
                        / "Hostx64"
                        / "x64"
                        / "cl.exe"
                    )
                    if not cl_path.exists():
                        typer.echo(f"cl.exe not found at {cl_path}")
                        raise FileNotFoundError
                    cl_location = str(cl_path)

                    found_compilers.append(
                        CompilerDetails(
                            name="msvc",
                            c_compiler=cl_location,
                            cpp_compiler=cl_location,
                        )
                    )
                else:
                    typer.echo(f"MSVC {version_str} found, but version 17+ is required")
        except (subprocess.SubprocessError, FileNotFoundError):
            typer.echo("MSVC not found")

        if not clang_valid and not msvc_valid:
            typer.echo("\nNo valid compiler found!")
            typer.echo("Please install at least one of the following:")
            typer.echo("  - Clang 19 or later")
            typer.echo("        can be installed from here https://releases.llvm.org/")
            typer.echo("  - MSVC 17 or later")
            typer.echo(
                "        can be installed from here https://visualstudio.microsoft.com/downloads/"
            )
            raise typer.Abort("Compiler validation failed")

        typer.echo("Compiler validation successful!")
        return found_compilers

    def generate_configuration_preset(
        self, compiler: CompilerDetails
    ) -> ConfigurePreset:
        if compiler.name == "msvc":
            return ConfigurePreset(
                name="msvc-toolchain",
                inherits=["base"],
                generator="Visual Studio 18 2026",
                # TODO choose correct visual studio based on whats installed
            )
        else:
            return ConfigurePreset(
                name="ninja-multi", inherits=["base"], generator="Ninja Multi-Config"
            )

    def get_script_extension(self) -> str:
        return "bat"

    def get_executable_extension(self) -> str:
        return ".exe"

    def _validate_dependencies(self) -> None:
        # Windows does not have any dependencies that need validating
        typer.echo("No dependencies to validate!")



import logging

import typer

from portal_tool.installer.repo.build_models import ConfigurePreset
from portal_tool.installer.configurators.configurator import (
    Configurator,
    CompilerDetails,
)


class MacConfigurator(Configurator):
    def __init__(self, yes: bool):
        logging.info("Running MacOs configurator")

    def _try_install_vcpkg_dependencies(self) -> None:
        pass

    def _install_package(self, packages: list[str]) -> None:
        raise NotImplementedError

    def validate_compilers(self) -> list[CompilerDetails]:
        typer.echo("Missing compiler validation, skipping...")
        return [
            CompilerDetails(name="clang", c_compiler="clang", cpp_compiler="clang++")
        ]

    def generate_configuration_preset(
        self, compiler: CompilerDetails
    ) -> ConfigurePreset:
        return ConfigurePreset(
            name="ninja-multi", inherits=["base"], generator="Ninja Multi-Config"
        )

    def get_script_extension(self) -> str:
        return "sh"

    def get_executable_extension(self) -> str:
        return ""

    def _validate_dependencies(self) -> None:
        typer.echo("No dependencies to validate!")



import json

import pathlib
import shutil

import typer

from portal_tool.framework.git_manager import GitManager
from portal_tool.models import Configuration, PortalModule, Dependency

IGNORED_FEATURES = ["dev"]


def get_portal_dependencies(depends_string: str) -> list[str]:
    if not depends_string.startswith("depends: "):
        typer.Abort(f"Invalid depends string: {depends_string}")

    cropped_string = depends_string[len("depends: ") :]
    return cropped_string.split(",")


class FrameworkManager:
    def __init__(self, config: Configuration):
        self.git_manager: GitManager = GitManager()
        self.git_manager.init_repo(config)

        self.framework_path = self.git_manager.data_folder / "repo" / "framework"

    def find_modules(self) -> list[PortalModule]:
        vcpkg_file = self.framework_path / "vcpkg.json"

        if not vcpkg_file.exists():
            typer.Abort("Could not find vcpkg.json in framework folder.")

        output = []
        vcpkg_data = json.loads(vcpkg_file.read_text())

        for name, feature in vcpkg_data.get("features", {}).items():
            if name in IGNORED_FEATURES:
                continue

            module_name = name if name.startswith("portal-") else f"portal-{name}"
            short_name = module_name.replace("portal-", "")
            description_tuple: tuple[str, str] | str = feature.get("description", "")
            dependencies: list[Dependency] = []
            has_portal_deps = False

            if isinstance(description_tuple, str):
                description = description_tuple
            else:
                description = description_tuple[0]
                for dep in get_portal_dependencies(description_tuple[1]):
                    dependencies.append(Dependency(name=dep))
                has_portal_deps = True

            for dependency in feature.get("dependencies", []):
                dependencies.append(
                    Dependency(
                        name=dependency.get("name"),
                        version=dependency.get("version>="),
                        features=dependency.get("features"),
                        platform=dependency.get("platform"),
                    )
                )

            options = ["PORTAL_BUILD_TESTS=OFF"]
            if has_portal_deps:
                options.append("PORTAL_FIND_PACKAGE=ON")

            output.append(
                PortalModule(
                    name=module_name,
                    short_name=short_name,
                    version=self.git_manager.get_version(short_name),
                    description=description,
                    options=options,
                    dependencies=dependencies,
                )
            )

        # TODO: validate portal dependencies version based on configures module versions
        return output

    def get_engine_version(self) -> str:
        return self.git_manager.get_version("engine")

    def get_vcpkg_baseline_hash(self) -> str:
        vcpkg_path = self.framework_path / "vcpkg.json"
        if not vcpkg_path.exists():
            typer.Abort("Could not find vcpkg.json in framework folder.")

        baseline = json.load(vcpkg_path.open())["builtin-baseline"]
        return baseline

    def list_examples(self) -> list[str]:
        examples_folder = self.framework_path / "examples"
        if not examples_folder.exists():
            return []

        return [
            example.name for example in examples_folder.iterdir() if example.is_dir()
        ]

    def configure_example(self, example_name: str, output_folder: pathlib.Path) -> None:
        example_folder = self.framework_path / "examples" / example_name
        if not example_folder.exists():
            typer.Abort(f"Example {example_name} not found.")

        output_folder.mkdir(parents=True, exist_ok=True)
        shutil.copytree(example_folder, output_folder, dirs_exist_ok=True)

        out_cmake = output_folder / "CMakeLists.txt"
        out_cmake_data = out_cmake.read_text()
        out_cmake_data = out_cmake_data.replace(
            "#find_package(portal-engine CONFIG REQUIRED)",
            "find_package(portal-engine CONFIG REQUIRED)",
        )
        out_cmake.write_text(out_cmake_data)

    def get_vcpkg_configuration(self) -> str:
        return (self.framework_path / "vcpkg-configuration.json").read_text()



import enum
import importlib
import importlib.util
import json
import os
import pathlib
import shlex
import subprocess
import zipimport
from typing import cast

import typer
from cookiecutter.main import cookiecutter

from portal_tool.installer.repo.build_models import (
    PackagePreset,
    WorkflowPreset,
    WorkflowStep,
)

from portal_tool.framework.framework_manager import FrameworkManager
from portal_tool.installer.configurators.configurator import CompilerDetails
from portal_tool.installer.configurator_factory import ConfiguratorFactory
from portal_tool.installer.repo.build_models import (
    CMakePresets,
    ConfigurePreset,
    BuildPreset,
)


class RepoMaker:
    """
    A repo consists of:
    - vcpkg (either submodule or globally installed)
    - src
    - resources
    - vcpkg configuration
    - cmake configuration
    """

    def __init__(self, path: pathlib.Path, framework_manager: FrameworkManager):
        self.configurator = ConfiguratorFactory().create(False)
        self.presets = CMakePresets()
        self.framework_manager = framework_manager

        self.name = typer.prompt("Project Name")
        self.base_path = pathlib.Path(
            typer.prompt(
                f"Base Location (will create a folder named {self.name.lower().replace(' ', '-')})",
                default=path,
            )
        )
        self.project_path = self.base_path / self.name.lower().replace(" ", "-")

        self.vcpkg_toolchain_location = "{}/scripts/buildsystems/vcpkg.cmake"
        self.use_global = False
        self.vcpkg_exec_location = pathlib.Path("")

        self._create_repo_from_template()

        use_example = typer.confirm(
            "Would you like to use an example project?", default=True
        )
        if use_example:
            available_examples = self.framework_manager.list_examples()

            example_choices = enum.Enum(
                "Examples", {ex: ex for ex in available_examples}
            )
            default_example = (
                "engine_test"
                if "engine_test" in available_examples
                else available_examples[0]
            )

            chosen_example = typer.prompt(
                f"Please choose an example to use ({', '.join(available_examples)})",
                type=example_choices,
                default=default_example,
            )

            self.framework_manager.configure_example(
                chosen_example.value, self.project_path
            )

        self._configure_git()
        self._setup_vcpkg()
        self._configure_build_system()
        self._update_vcpkg_baseline()

    def _find_pacakge_path(self) -> pathlib.Path:
        source_path = "templates/repo/bootstrap"
        spec = importlib.util.find_spec("portal_tool")
        assert spec is not None, "An import spec was not found for the package."
        loader = spec.loader
        assert loader is not None, "A loader was not found for the package."

        if isinstance(loader, zipimport.zipimporter):
            self._archive = loader.archive
            pkgdir = next(iter(spec.submodule_search_locations))  # type: ignore
            template_root = os.path.join(pkgdir, source_path).rstrip(os.path.sep)
        else:
            roots: list[str] = []

            # One element for regular packages, multiple for namespace
            # packages, or None for single module file.
            if spec.submodule_search_locations:
                roots.extend(spec.submodule_search_locations)
            # A single module file, use the parent directory instead.
            elif spec.origin is not None:
                roots.append(os.path.dirname(spec.origin))

            if not roots:
                raise ValueError(
                    "The portal_tool package was not installed in a"
                    " way that PackageLoader understands."
                )

            for root in roots:
                root = os.path.join(root, source_path)

                if os.path.isdir(root):
                    template_root = root
                    break
            else:
                raise ValueError(
                    f"PackageLoader could not find a {source_path!r} directory"
                    f" in the portal_tool package."
                )
        return pathlib.Path(template_root)

    def _create_repo_from_template(self) -> None:
        if self.project_path.exists():
            proceed = typer.confirm(
                "Found existing project, would you like to continue? (delete folder)"
            )
            if not proceed:
                raise typer.Abort("Aborting repo creation.")
            return

        cookiecutter(
            self._find_pacakge_path().as_posix(),
            no_input=True,
            extra_context={
                "project_name": self.name,
                "engine_version": self.framework_manager.get_engine_version(),
                "vcpkg_baseline": self.framework_manager.get_vcpkg_baseline_hash(),
            },
            output_dir=self.base_path.as_posix(),
        )

    def _configure_git(self) -> None:
        typer.echo(f"Initializing git repo in: {self.project_path}")
        subprocess.check_output(shlex.split(f'git -C "{self.project_path}" init'))

    def _setup_vcpkg(self) -> None:
        vcpkg_root, found_using_env = self.configurator.find_vcpkg_root()

        self.use_global = False
        if vcpkg_root:
            self.use_global = typer.confirm(
                f"Found global vcpkg, [{vcpkg_root}] would you like to use? (if not, a local submodule will be created)",
                default=True,
            )

        if not self.use_global:
            typer.echo(f"Creating vcpkg submodule in: {self.project_path / 'vcpkg'}")
            try:
                subprocess.check_output(
                    shlex.split(
                        f'git -C "{self.project_path}" submodule add https://github.com/microsoft/vcpkg "{"vcpkg"}"'
                    )
                )
                subprocess.check_output(
                    f"{self.project_path.as_posix()}/vcpkg/bootstrap-vcpkg.{self.configurator.get_script_extension()}",
                    shell=True,
                )
                self.vcpkg_exec_location = (
                    self.project_path
                    / "vcpkg"
                    / f"vcpkg{self.configurator.get_executable_extension()}"
                )
            except subprocess.CalledProcessError:
                typer.echo("Failed to add submodule, please add it manually.")
        else:
            self.vcpkg_exec_location = (
                cast(pathlib.Path, vcpkg_root)
                / f"vcpkg{self.configurator.get_executable_extension()}"
            )

        self.vcpkg_toolchain_location = self.vcpkg_toolchain_location.format(
            (
                "$env{VCPKG_ROOT}"
                if found_using_env
                else (pathlib.Path(os.path.expanduser("~")) / ".vcpkg").as_posix()
            )
            if self.use_global
            else "${sourceDir}/vcpkg"
        )

    def _configure_build_system(self) -> None:
        project_compiler: CompilerDetails
        compilers = self.configurator.validate_compilers()
        if len(compilers) == 0:
            raise typer.Abort(
                "No valid compilers found, please run `portal-tool install` and follow the instructions."
            )

        if len(compilers) == 1:
            project_compiler = compilers[0]

        else:
            choices_data = {comp.name: comp.name for comp in compilers}
            compilers_choices = enum.Enum("Compilers", choices_data)

            default_name = "clang" if "clang" in choices_data else compilers[0].name

            choice_member = typer.prompt(
                f"Multiple compilers found, please choose one ({', '.join(choices_data.keys())})",
                type=compilers_choices,
                default=default_name,
            )

            comp_lookup = {comp.name: comp for comp in compilers}
            project_compiler = comp_lookup[choice_member.value]

        base = ConfigurePreset(
            name="base",
            hidden=True,
            binary_dir="${sourceDir}/build/${presetName}",
            cache_variables={
                "CMAKE_TOOLCHAIN_FILE": self.vcpkg_toolchain_location,
                "CMAKE_CONFIGURATION_TYPES": "Debug;RelWithDebInfo;Release",
                "CMAKE_C_COMPILER": project_compiler.c_compiler,
                "CMAKE_CXX_COMPILER": project_compiler.cpp_compiler,
            },
        )
        base.environment = {
            "PORTAL_C_COMPILER": project_compiler.c_compiler,
            "PORTAL_CPP_COMPILER": project_compiler.cpp_compiler,
            "VCPKG_KEEP_ENV_VARS": "PORTAL_C_COMPILER;PORTAL_CPP_COMPILER;PATH",
        }

        generator = self.configurator.generate_configuration_preset(project_compiler)

        self.presets.configure_presets = [base, generator]

        self.presets.build_presets = [
            BuildPreset(
                name="debug", configure_preset=generator.name, configuration="Debug"
            ),
            BuildPreset(
                name="development",
                configure_preset=generator.name,
                configuration="RelWithDebInfo",
            ),
            BuildPreset(
                name="dist",
                configure_preset=generator.name,
                configuration="Release",
            ),
        ]

        self.presets.package_presets = [
            PackagePreset(
                name="pack-zip",
                configure_preset=generator.name,
                package_directory="${sourceDir}/dist",
                generators=["ZIP"],
            ),
            PackagePreset(
                name="pack-installer",
                configure_preset=generator.name,
                package_directory="${sourceDir}/dist",
                generators=["IFW"],
            ),
        ]

        self.presets.workflow_presets = [
            WorkflowPreset(
                name="package-zip",
                steps=[
                    WorkflowStep(name=generator.name, type="configure"),
                    WorkflowStep(name="dist", type="build"),
                    WorkflowStep(name="pack-zip", type="package"),
                ],
            ),
            WorkflowPreset(
                name="package-installer",
                steps=[
                    WorkflowStep(name=generator.name, type="configure"),
                    WorkflowStep(name="dist", type="build"),
                    WorkflowStep(name="pack-installer", type="package"),
                ],
            ),
        ]

        (self.project_path / "CMakePresets.json").write_text(
            self.presets.model_dump_json(indent=4, exclude_none=True, by_alias=True)
        )

        (self.project_path / "vcpkg-configuration.json").write_text(
            self.framework_manager.get_vcpkg_configuration()
        )

    def _update_vcpkg_baseline(self) -> None:
        output = subprocess.check_output(
            shlex.split(
                f"{self.vcpkg_exec_location.as_posix()} x-update-baseline --dry-run"
            ),
            cwd=self.project_path.as_posix(),
        )

        portal_vcpkg_baseline = None
        output_lines = output.decode().splitlines()
        for line in output_lines:
            if "https://github.com/JonatanNevo/portal-vcpkg-registry" in line:
                portal_vcpkg_baseline = line.split(" ")[-1].strip("'")

        if portal_vcpkg_baseline is None:
            typer.echo(
                "Failed to update portal-vcpkg-registry baseline, please make sure it points the correct commit."
            )
            return

        vcpkg_configuration = json.loads(
            (self.project_path / "vcpkg-configuration.json").read_text()
        )
        for registry in vcpkg_configuration["registries"]:
            if (
                registry["repository"]
                == "https://github.com/JonatanNevo/portal-vcpkg-registry"
            ):
                registry["baseline"] = portal_vcpkg_baseline
        json.dump(
            vcpkg_configuration,
            (self.project_path / "vcpkg-configuration.json").open("w"),
            indent=2,
        )



import logging
import pathlib
from typing import Annotated
from importlib.metadata import version as meta_version

import typer

from portal_tool.runner import runner
from portal_tool.framework.framework_manager import FrameworkManager
from portal_tool.installer.repo.repo_maker import RepoMaker
from portal_tool.installer.installer import Installer
from portal_tool.models import Configuration
from portal_tool.registry.registry_update import RegistryManager

global_working_directory = pathlib.Path.cwd()
configuration: Configuration
framework_manager: FrameworkManager


def version_callback(value: bool):
    if value:
        typer.echo(f"Portal Tool version: {meta_version('portal_tool')}")
        raise typer.Exit()


def registry_start(
    working_dir: Annotated[
        pathlib.Path,
        typer.Option(
            "-d", "--working-dir", help="The working directory of the project"
        ),
    ] = pathlib.Path.cwd(),
):
    global global_working_directory
    global_working_directory = working_dir


registry = typer.Typer(callback=registry_start)


@registry.command()
def update() -> None:
    manager = RegistryManager(global_working_directory, framework_manager)
    manager.update()


@registry.command()
def update_versions() -> None:
    manager = RegistryManager(global_working_directory, framework_manager)
    manager.update_version()


app = typer.Typer()
app.add_typer(registry, name="registry", help="Commands for managing the registry")
app.add_typer(runner, name="run", help="Commands for running the project")


@app.command()
def install(
    dependencies_only: Annotated[
        bool,
        typer.Option(
            "--only-dependencies",
            help="Only installs the dependencies, skips all other stages",
        ),
    ] = False,
    auto_yes: Annotated[
        bool,
        typer.Option(
            "-y",
            "--yes",
            help="Automatically answer yes to all questions.",
        ),
    ] = False,
) -> None:
    installer = Installer(
        configuration.registry_url_template.format(
            repo=configuration.vcpkg_registry_repo
        ),
        auto_yes,
    )

    if dependencies_only:
        installer.install(False, True)
    else:
        installer.install(True, True)


@app.command()
def init(
    directory: Annotated[
        pathlib.Path,
        typer.Option("-d", "--dir", help="The directory to initialize the project in"),
    ] = pathlib.Path.cwd(),
) -> None:
    RepoMaker(directory, framework_manager)


@app.callback()
def main(
    version: Annotated[
        bool | None,
        typer.Option(
            "--version",
            "-v",
            callback=version_callback,
            is_eager=True,
            help="Show the version and exit.",
        ),
    ] = None,
    framework_repo: Annotated[
        str,
        typer.Option("--framework", help="The repo of portal-framework"),
    ] = "JonatanNevo/portal-framework",
    working_branch: Annotated[
        str,
        typer.Option("--branch", help="The branch of the framework repo to work on"),
    ] = "main",
    registry_repo: Annotated[
        str, typer.Option("--registry", help="The repo of the vcpkg registry")
    ] = "JonatanNevo/portal-vcpkg-registry",
):
    global configuration
    global framework_manager
    logging.basicConfig(level=logging.INFO)

    configuration = Configuration(
        repo=framework_repo,
        repo_branch=working_branch,
        vcpkg_registry_repo=registry_repo,
    )

    framework_manager = FrameworkManager(configuration)


if __name__ == "__main__":
    app()



import json
import shlex
import subprocess
from enum import Enum
from typing import Annotated
import re

import pathlib
import typer

from portal_tool.installer.configurator_factory import ConfiguratorFactory

global_working_directory = pathlib.Path.cwd()


class Configuration(str, Enum):
    debug = "debug"
    development = "development"
    dist = "dist"


CMAKE_CONFIGURATION: dict[Configuration, str] = {
    Configuration.debug: "Debug",
    Configuration.development: "RelWithDebInfo",
    Configuration.dist: "Release",
}

runner = typer.Typer()


def find_configure_preset_name(working_dir: pathlib.Path) -> str:
    presets_path = working_dir / "CMakePresets.json"
    if not presets_path.exists():
        raise typer.Abort("CMakePresets.json not found in the working directory.")

    presets = json.loads(presets_path.read_text())
    for preset in presets.get("configurePresets", []):
        if not preset.get("hidden", False):
            return preset["name"]

    raise typer.Abort("No visible configure preset found in CMakePresets.json.")


def first_cmake_call_arg(text: str, func_name: str) -> str | None:
    """
    Extract the first argument of: func_name(<arg> ...)
    Handles arbitrary whitespace/newlines and quoted first args.
    """
    # Match: func_name ( <spaces> ( "quoted" | bareword )
    pattern = rf"""
        ^\s*{re.escape(func_name)}\s*    # function name at line start (with indentation)
        \(\s*                           # opening paren, then whitespace/newlines
        (                               # capture group 1 = first arg
            "(?:\\.|[^"\\])*"           # "quoted string" with simple escapes
            |
            [^\s\)]+                    # or bare token up to whitespace or ')'
        )
    """
    m = re.search(pattern, text, re.MULTILINE | re.DOTALL | re.VERBOSE)
    if not m:
        return None

    arg = m.group(1)
    if arg.startswith('"') and arg.endswith('"'):
        arg = arg[1:-1]  # unquote (optional; remove if you want raw token)
    return arg


def find_target_name(cmake_path: pathlib.Path) -> str:
    target_name = first_cmake_call_arg(cmake_path.read_text(), "portal_add_game")
    if target_name is None:
        return typer.prompt(
            "Unable to find target name in CMakeLists.txt. Please enter it manually:"
        )
    return target_name


class Runner:
    def __init__(
        self,
        working_dir: pathlib.Path,
        configuration: Configuration,
        build_path: pathlib.Path | None,
        cmake_configuration: str | None,
    ):
        self.working_dir = working_dir
        self.configuration = configuration
        self.configurator = ConfiguratorFactory().create(False)
        self.preset_name = find_configure_preset_name(working_dir)

        self.cmake_configuration = cmake_configuration
        if self.cmake_configuration is None:
            self.cmake_configuration = CMAKE_CONFIGURATION[configuration]

        if build_path is None:
            build_path = pathlib.Path("build") / self.preset_name

        self.target_name = find_target_name(working_dir / "CMakeLists.txt")
        self.executable_dir = working_dir / build_path / self.cmake_configuration

    def build_executable(self) -> None:
        subprocess.run(
            shlex.split(f"cmake --fresh --preset {self.preset_name}"),
            cwd=self.working_dir,
            check=True,
        )

        subprocess.run(
            shlex.split(f"cmake --build --preset {self.configuration.name}"),
            cwd=self.working_dir,
            check=True,
        )

    def validate_executable(self, executable: str, clean: bool) -> None:
        exe_path = pathlib.Path(executable)
        if clean or not exe_path.exists():
            typer.confirm(
                "No executable found. Would you like to build it now?", abort=True
            )

            self.build_executable()

    def run(self, *args: str) -> None:
        subprocess.Popen([*args], cwd=self.executable_dir)


@runner.command(name="editor")
def run_editor(
    working_dir: Annotated[
        pathlib.Path,
        typer.Option(
            "-d", "--working-dir", help="The working directory of the project"
        ),
    ] = pathlib.Path.cwd(),
    configuration: Annotated[
        Configuration,
        typer.Option("-c", "--configuration", help="Which configuration to use/build"),
    ] = Configuration.development,
    build_path: Annotated[
        pathlib.Path | None,
        typer.Option(
            "-b", "--build-path", help="The path to the build directory, (relative)"
        ),
    ] = None,
    cmake_configuration: Annotated[
        str | None,
        typer.Option(
            "--configuration-build-path",
            help="The path in the build directory for the configuration",
        ),
    ] = None,
    clean: Annotated[
        bool,
        typer.Option("--clean", help="Always rebuild the executable before running"),
    ] = False,
):
    engine_runner = Runner(working_dir, configuration, build_path, cmake_configuration)
    executable = f"{engine_runner.executable_dir / engine_runner.target_name}_editor{engine_runner.configurator.get_executable_extension()}"
    engine_runner.validate_executable(executable, clean=clean)

    typer.echo(f"Running editor from: {executable}")
    engine_runner.run(executable, "-p", f"{working_dir}")


@runner.command(name="runtime")
def run_runtime(
    working_dir: Annotated[
        pathlib.Path,
        typer.Option(
            "-d", "--working-dir", help="The working directory of the project"
        ),
    ] = pathlib.Path.cwd(),
    configuration: Annotated[
        Configuration,
        typer.Option("-c", "--configuration", help="Which configuration to use/build"),
    ] = Configuration.development,
    build_path: Annotated[
        pathlib.Path | None,
        typer.Option("-b", "--build-path", help="The path to the build directory"),
    ] = None,
    cmake_configuration: Annotated[
        str | None,
        typer.Option(
            "--configuration-build-path",
            help="The path in the build directory for the configuration",
        ),
    ] = None,
    clean: Annotated[
        bool,
        typer.Option("--clean", help="Always rebuild the executable before running"),
    ] = False,
):
    engine_runner = Runner(working_dir, configuration, build_path, cmake_configuration)
    executable = f"{engine_runner.executable_dir / engine_runner.target_name}{engine_runner.configurator.get_executable_extension()}"
    engine_runner.validate_executable(executable, clean=clean)

    typer.echo(f"Running game from: {executable}")
    engine_runner.run(executable)



from pydantic import BaseModel, Field
from pydantic_settings import BaseSettings


class GitDetails(BaseModel):
    repo: str
    ref: str
    sha: str
    head_ref: str
    subdirectory: str


class Dependency(BaseModel):
    name: str
    version: str | None = None
    features: list[str] | None = None
    platform: str | None = None


class Feature(BaseModel):
    name: str
    description: str
    dependencies: list[Dependency] = Field(default_factory=list)


class PortalModule(BaseModel):
    name: str
    short_name: str = ""
    version: str = ""
    description: str = ""
    options: list[str] = Field(default_factory=list)
    dependencies: list[Dependency] = Field(default_factory=list)


class GitStatus(BaseModel):
    repo: str | None = None
    branch: str | None = None
    commit: str | None = None
    sha: str | None = None


class LocalStatus(BaseModel):
    framework: GitStatus = Field(default_factory=GitStatus)
    registry: GitStatus = Field(default_factory=GitStatus)


class Configuration(BaseSettings):
    repo: str = "JonatanNevo/portal-framework"
    repo_branch: str = "main"
    vcpkg_registry_repo: str = "JonatanNevo/portal-vcpkg-registry"
    registry_url_template: str = "github.com:{repo}"



import json
import logging
import pathlib
import shlex
import subprocess

import jinja2

from portal_tool.framework.framework_manager import FrameworkManager
from portal_tool.framework.git_manager import GitManager
from portal_tool.models import GitDetails, PortalModule


def filter_nones(obj):
    """
    Given a JSON-serializable object, return a copy without elements which have
    a value of None.
    """

    if isinstance(obj, list):
        # This version may or may not be easier to read depending on your
        # preferences:
        # return list(filter(None, map(remove_nones, obj)))

        # This version uses a generator expression to avoid computing a full
        # list only to immediately walk it again:
        filtered_values = (filter_nones(j) for j in obj)
        return [i for i in filtered_values if i is not None]
    elif isinstance(obj, dict):
        filtered_items = ((i, filter_nones(j)) for i, j in obj.items())
        return {k: v for k, v in filtered_items if v is not None}
    else:
        return obj


class RegistryManager:
    def __init__(
        self, registry_path: pathlib.Path, framework_manager: FrameworkManager
    ):
        self.registry_path = registry_path
        self.framework_manager = framework_manager

    def update(self) -> None:
        logging.info(f"Updating vcpkg registry at: {self.registry_path.absolute()}")

        git_manager: GitManager = GitManager()
        for module in self.framework_manager.find_modules():
            details = git_manager.to_details(subdirectory=module.short_name)
            self._generate_vcpkg_port(module, details)

    def update_version(self) -> None:
        output = subprocess.check_output(
            shlex.split(
                f'vcpkg --x-builtin-ports-root="{self.registry_path / "ports"}" --x-builtin-registry-versions-dir="{self.registry_path / "versions"}" x-add-version --all --verbose'
            )
        )
        print(output.decode())

    def _generate_vcpkg_port(
        self, module: PortalModule, git_details: GitDetails
    ) -> None:
        env = jinja2.environment.Environment(
            loader=jinja2.PackageLoader("portal_tool", "templates/vcpkg"),
        )

        cmake_template = env.get_template("portfile.cmake.j2")

        cmake = cmake_template.render(module=module, git=git_details)
        usage_template = env.get_template("usage.j2")
        usage = usage_template.render(module=module, git=git_details)

        port_path = self.registry_path / "ports" / module.name
        if not port_path.exists():
            port_path.mkdir(parents=True)

        logging.info(f"Generating vcpkg port at: {port_path}")

        port_file = port_path / "portfile.cmake"
        vcpkg_file = port_path / "vcpkg.json"
        usage_file = port_path / "usage"

        port_file.write_text(cmake)
        self._make_vcpkg_json(module, vcpkg_file)
        usage_file.write_text(usage)

        subprocess.check_output(shlex.split(f'vcpkg format-manifest "{vcpkg_file}"'))

    @staticmethod
    def _make_vcpkg_json(module: PortalModule, json_path: pathlib.Path) -> None:
        old_vcpkg_json = json.loads(
            json_path.read_text() if json_path.exists() else "{}"
        )

        json_details = {
            "name": module.name,
            "version": module.version,
            "description": module.description,
            "license": "MIT",
        }

        json_deps = [
            {"name": "vcpkg-cmake", "host": True},
            {"name": "vcpkg-cmake-config", "host": True},
        ]
        for dependency in module.dependencies:
            json_deps.append(
                {
                    "name": dependency.name,
                    "features": dependency.features,
                    "version>=": dependency.version,
                    "platform": dependency.platform,
                }
            )
        json_details["dependencies"] = json_deps

        # json_features = {}
        # for feature in module.features:
        #     json_feature = {
        #         "description": feature.description,
        #         "dependencies": []
        #     }
        #
        #     for dep in feature.dependencies:
        #         json_feature["dependencies"].append(
        #             {
        #                 "name": dep.name,
        #                 "features": dep.features,
        #                 "version>=": dep.version,
        #                 "platform": dep.platform
        #             }
        #         )
        #     json_features[feature.name] = json_feature

        if json_details["version"] == old_vcpkg_json.get("version"):
            json_details["port-version"] = old_vcpkg_json.get("port-version", 0) + 1

        formatted_json = filter_nones(json_details)
        json_path.write_text(json.dumps(formatted_json, indent=4))
